bell = """OPENQASM 3.0;
include "qelib1.inc";
qreg q[2];
creg c[2];
h q[0];
cx q[0],q[1];
measure q[0] -> c[0];
measure q[1] -> c[1];"""

ghz = """OPENQASM 3.0;
include "qelib1.inc";
qreg q[3];
creg meas[3];
h q[0];
cx q[0],q[1];
cx q[0],q[2];
barrier q[0],q[1],q[2];
measure q[0] -> meas[0];
measure q[1] -> meas[1];
measure q[2] -> meas[2];"""



from qmio.backends import QPUBackend

class QmioRuntimeService():

    def backend(self, name):
        if name == "qpu":
            return QPUBackend()
        # if name == "qulacs":
        #     pass
        else:
            raise ValueError(f"Backend unknown: {name}")



import pytest
from qmio.services import QmioRuntimeService

@pytest.fixture
def runtime_service():
    return QmioRuntimeService()


def test_backend_selection(runtime_service):
    backend = runtime_service.backend(name="qpu")
    assert isinstance(backend, QPUBackend)


def test_invalid_backend_name(runtime_service):
    with pytest.raises(ValueError):
        runtime_service.backend(name="invalidBackend")



### Tests de las funciones del backend qpu ###
import pytest
from config.development import ZMQ_SERVER
from qmio.qmio import circuit
from qmio.backends import _config_build
import zmq


def test_config_build():
    result_config = _config_build(shots=100)
    assert result_config == '{"$type": "<class \'qat.purr.compiler.config.CompilerConfig\'>", "$data": {"repeats": 100, "repetition_period": null, "results_format": {"$type": "<class \'qat.purr.compiler.config.QuantumResultsFormat\'>", "$data": {"format": {"$type": "<enum \'qat.purr.compiler.config.InlineResultsProcessing\'>", "$value": 1}, "transforms": {"$type": "<enum \'qat.purr.compiler.config.ResultsFormatting\'>", "$value": 3}}}, "metrics": {"$type": "<enum \'qat.purr.compiler.config.MetricsType\'>", "$value": 6}, "active_calibrations": [], "optimizations": null}}'



from qmio.clients import ZMQClient
import json


def _optimization_options_builder(optimization, opt_type="Tket"):
    """
    Optimization builder to not depend on qat

    Just Tket for now
    """
    if opt_type != "Tket":
        raise TypeError(f"{opt_type}: Not a valid type")
    if optimization == 1:
        opt_value = 18
    elif optimization == 2:
        opt_value = 30
    elif optimization == 0:
        opt_value = 1
    else:
        raise ValueError(f"{optimization}: Not a valid Optimization Value")
    return opt_value


def _results_format_builder(res_format="binary_count"):
    """
    Results Format Builder without QAT

    Posibilities:
    raw:

    binary:
    Binary string
    binary_count (default):
    Returns a count of each instance of measured qubit registers.
        Switches result format to raw.
    squash_binary_result_arrays:
    Squashes binary result list into a singular bit string. Switches results to
        binary.
    """
    if res_format != "binary_count":
        if res_format == "raw":
            # Check
            InlineResultsProcessing = 1
            ResultsFormatting = 2
        elif res_format == "binary":
            # Check
            InlineResultsProcessing = 2
            ResultsFormatting = 2
        elif res_format == "squash_binary_result_arrays":
            # Check
            InlineResultsProcessing = 2
            ResultsFormatting = 6
    elif res_format == "binary_count":
        # Desde recién instanciado - check
        InlineResultsProcessing = 1
        ResultsFormatting = 3
    else:
        raise TypeError(f"{res_format}: Not a valid result format")
    return InlineResultsProcessing, ResultsFormatting


def _config_build(shots: int, repetition_period=None, optimization=0, res_format="binary_count"):
    """
    Builds a config json object from options
    not qat dependent
    """
    inlineResultsProcessing, resultsFormatting = _results_format_builder(res_format)
    opt_value = _optimization_options_builder(optimization=optimization)
    config = {
        "$type": "<class \'qat.purr.compiler.config.CompilerConfig\'>",
        "$data": {
            "repeats": shots,
            "repetition_period": repetition_period,
            "results_format": {
                "$type": "<class \'qat.purr.compiler.config.QuantumResultsFormat\'>",
                "$data": {
                    "format": {
                        "$type": "<enum \'qat.purr.compiler.config.InlineResultsProcessing\'>",
                        "$value": inlineResultsProcessing
                    },
                    "transforms": {
                        "$type": "<enum \'qat.purr.compiler.config.ResultsFormatting\'>",
                        "$value": resultsFormatting
                    }
                }
            },
            "metrics": {
                "$type": "<enum \'qat.purr.compiler.config.MetricsType\'>",
                "$value": 6
            },
            "active_calibrations": [],
            "optimizations": {
                "$type": "<enum \'qat.purr.compiler.config.TketOptimizations\'>",
                "$value": opt_value
            }
        }
    }
    config_str = json.dumps(config)
    return config_str


class QPUBackend:
    def __init__(self):
        self.client = None

    def __enter__(self):
        self.client = ZMQClient()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.client:
            self.client.close()
            self.client = None

    def connect(self):
        self.client = ZMQClient()

    def disconnect(self):
        self.client.close()
        self.client = None

    # def run_v0(self, circuit, config):
    #     if not self.client:
    #         raise RuntimeError("Not connected to the server")

    #     job = (circuit, config)
    #     self.client._send(job)
    #     result = self.client._await_results()
    #     return result

    # def run_v1(self, circuit, shots):
    #     if not self.client:
    #         raise RuntimeError("Not connected to the server")

    #     config = _config_build(shots)
    #     job = (circuit, config)
    #     self.client._send(job)
    #     result = self.client._await_results()
    #     return result

    def run(self, circuit, shots, repetition_period=None, optimization=0, res_format="binary_count"):
        if not self.client:
            raise RuntimeError("Not connected to the server")

        config = _config_build(shots, repetition_period=repetition_period, optimization=optimization, res_format=res_format)
        job = (circuit, config)
        self.client._send(job)
        result = self.client._await_results()
        return result



from time import time
from typing import Union
from config import ZMQ_SERVER
import zmq

class ZMQBase:
    def __init__(self, socket_type):
        self._context = zmq.Context()
        self._socket = self._context.socket(socket_type)
        self._timeout = 30.0
        #self._address = "tcp://10.133.29.226:5556"
        self._address = ZMQ_SERVER

    def _check_recieved(self):
        try:
            msg = self._socket.recv_pyobj()
            return msg
        except zmq.ZMQError:
            return None

    def _send(self, message) -> None:
        sent = False
        t0 = time()
        while not sent:
            try:
                self._socket.send_pyobj(message)
                sent = True
            except zmq.ZMQError as e:
                if time() > t0 + self._timeout:
                    raise TimeoutError(
                        "Sending %s on %s timedout" % (message, self._address)
                    )
        return

    def close(self):
        """Disconnect the link to the socket."""
        if self._socket.closed:
            return
        self._socket.close()
        self._context.destroy()

    def __del__(self):
        self.close()


class ZMQClient(ZMQBase):
    def __init__(self):
        super().__init__(zmq.REQ)
        self._socket.setsockopt(zmq.LINGER, 0)
        self._socket.connect(self._address)

    def _await_results(self):
        result = None
        while result is None:
            result = self._check_recieved()
        return result



from qmio.services import QmioRuntimeService

circuit = """OPENQASM 2.0;
    include "qelib1.inc";
    qreg q[4];
    creg c[4];
    rx(0.01) q[0];
    measure q[0]->c[0];
    measure q[1]->c[1];
    measure q[2]->c[2];
    measure q[3]->c[3];
    """

config = '{"$type": "<class \'qat.purr.compiler.config.CompilerConfig\'>", "$data": {"repeats": 1, "repetition_period": null, "results_format": {"$type": "<class \'qat.purr.compiler.config.QuantumResultsFormat\'>", "$data": {"format": {"$type": "<enum \'qat.purr.compiler.config.InlineResultsProcessing\'>", "$value": 1}, "transforms": {"$type": "<enum \'qat.purr.compiler.config.ResultsFormatting\'>", "$value": 3}}}, "metrics": {"$type": "<enum \'qat.purr.compiler.config.MetricsType\'>", "$value": 6}, "active_calibrations": [], "optimizations": null}}'

# service = QmioRuntimeService()

# backend = service.backend(name="qpu")
# backend.connect()
# result = backend.run(program, config)
# backend.disconnect()

# with service.backend(name="qpu") as backend:
#     result = backend.run(program=program, config=config)

# print(result)
